/*
	Gera pginas de documentao - by Juci, em 26 de julho 1998.
	Inspirado no javadoc, da Sun.

	Este programa  software livre. Voc pode usar e alterar  vontade,
	sem pedir autorizao a ningum, s que suas alteraes tambm devem
	ser compartilhadas, de modo que voc no pode restringir o uso delas.
	Voc pode incluir mais informaes no rodap das pginas geradas, mas
	no pode tirar o link para a pgina oficial do GERADOC.
	Este comentrio deve acompanhar todas as verses do programa.

	A verso oficial, que agrega todas as melhorias,  mantida em
	http://casa.dos.programadores.com.br/

	Para incluir suas alteraes na verso oficial mande-as num e-mail para
	casa@programadores.com.br

*/
#include <ctype.h>
#include <iostream.h>
#include <fstream.h>
#include <windows.h>
#include <vector>
#include <algorithm>
#include <string>
#include <stdlib.h>

/* Ative a linha abaixo para incluir membros private nas descries */
#define EXIBIR_MEMBROS_PRIVADOS

typedef std::string String;

/**
	Esta classe  a interface com o arquivo HTML que est sendo gerado.
	Atravs dela que  feita a gravao do texto e das tags.
*/
class GeradorDeHtml
{
	/**
		Para permitir: ob << nome;
	*/
	friend GeradorDeHtml &operator<<(GeradorDeHtml &lhs, String &rhs);

	/**
		Para permitir: ob << "Literal";
	*/
	friend GeradorDeHtml &operator<<(GeradorDeHtml &lhs, const char *rhs);

	/**
		Objeto utilizado para gravar as informaes no arquivo texto.
	*/
	ofstream arq;

public:
	/**
		Para criar um objeto deste voc passa o ttulo da pgina
		e o nome do arquivo.
	*/
	GeradorDeHtml(String &titulo, String &nomeDoArquivo);
	~GeradorDeHtml();

	/**
		Aumenta o nvel de endentao.
	*/
	void operator++();

	/**
		Diminui o nvel de endentao.
	*/
	void operator--();

	/**
		Gera um link hipertexto para a pgina.
		No  necessrio incluir a extenso html no nome da pgina.
	*/
	void emiteLink(String &texto, String &nomeDaPagina);

	/**
		Isola o texto passado na string "s" num pargrafo. Opcionalmente
		voc pode especificar o nome da fonte a ser utilizada.
	*/
	void paragrafo(String &s, bool bold = false);

	/**
		Permite que voc abra um pargrafo explicitamente.
	*/
	void abreParagrafo();

	/**
		Usado para fechar um pargrafo que foi aberto com abreParagrafo().
	*/
	void fechaParagrafo();

	/**
		Fora uma quebra para a linha seguinte.
	*/
	void novaLinha();

	/**
		Emite em ttulo em letras garrafais.
	*/
	void garrafais(String &);
};


GeradorDeHtml &operator<<(GeradorDeHtml &lhs, String &rhs)
{
	lhs.arq << rhs.c_str();
	return lhs;
}


GeradorDeHtml &operator<<(GeradorDeHtml &lhs, const char *rhs)
{
	lhs.arq << rhs;
	return lhs;
}

void GeradorDeHtml::operator++()
{
	arq << "<DIR>";
}

void GeradorDeHtml::operator--()
{
	arq << "</DIR>";
}

void GeradorDeHtml::novaLinha()
{
	++(*this);
	--(*this);
}

void GeradorDeHtml::abreParagrafo()
{
	arq << "<P CLASS=\"t\">";
}

void GeradorDeHtml::fechaParagrafo()
{
	arq << "</P>";
}

void GeradorDeHtml::paragrafo(String &s, bool bold)
{
	if(bold)
		arq << "<B>";

	abreParagrafo();
	arq << s.c_str();
	fechaParagrafo();
	if(bold)
		arq << "</B>";
}


void GeradorDeHtml::garrafais(String &texto)
{
	arq << "<H1 CLASS=\"h1\">" << texto.c_str() << "</H1>";
}


GeradorDeHtml::GeradorDeHtml(String &titulo, String &nome)
: arq((nome + ".html").c_str())
{
	arq << "<HTML><FONT FACE=\"Verdana\">\n"
		"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=iso-8859-1\">"
		"<META NAME=\"Author\" CONTENT=\"Juci Dias Andrade\">"
		"<META NAME=\"GENERATOR\" CONTENT=\"Mozilla/4.03 [pt] (WinNT; I) [Netscape]\">"
		"<HEAD><TITLE>" << titulo.c_str() << "</TITLE>"
		"</HEAD><BODY BACKGROUND=\"fundo.jpg\">";
}


void GeradorDeHtml::emiteLink(String &texto, String &nomeDaPagina)
{
	arq << "<A class=\"normal\" href=\"" << nomeDaPagina.c_str() << ".html\">"
		<< texto.c_str() << "</A>";
}

GeradorDeHtml::~GeradorDeHtml()
{
	arq <<
		"<P><DL><HR WIDTH=\"100%\">"
		"<FONT SIZE=-2>Esta documenta&ccedil;&atilde;o "
		"foi gerada automaticamente com a ferramenta <A HREF=\"http://casa.dos.programadores.com.br/\">"
		"GERADOC</A>.</FONT></DL></P>"
		"</BODY></FONT></HTML>\n";
}

/**
	Esta classe permite que se leia um arquivo colocando linha a linha em
	Strings. Voc tambm pode pegar o nmero da linha atual.
*/
class StreamNumerado
{
	/**
		Objeto utilizado para ler o arquivo texto.
	*/
	istream &s;

	/**
		Armazena o nmero da ltima linha lida.
	*/
	int linhaAtual;

	/**
		Buffer interno que contm a linha lida.
	*/
	char *buffer;

	/**
		Indica se deve reler a ltima linha lida.
	*/
	bool linhaFoiDevolvida;
public:

	/**
		Tamanho mximo de uma linha na declarao de classe.
	*/
	enum { TAMBUF = 1024 };

	/**
		Para criar uma instncia voc passa uma referncia a um
		istream que j exista, por exemplo um fstream recm criado.
	*/
	StreamNumerado(istream &s_)
		: s(s_), linhaAtual(0), buffer(new char[TAMBUF]), linhaFoiDevolvida(false)
	{}

	~StreamNumerado() { delete buffer; }

	/**
		Cada chamada a este mtodo devolve uma linha do arquivo, j sem
		espaos no incio e no fim.
	*/
	String pegaLinha();

	/**
		Extrai um caracter do fluxo de entrada.
	*/
	char get() { return s.get(); }

	/**
		Devolve o prximo caracter do fluxo de entrada sem extra-lo.
	*/
	char peek() { return s.peek(); }

	/**
		Devolve true se o fluxo chegou ao fim.
	*/
	bool acabou() { return s.eof() != 0; }

	/**
		Nmero da ltima linha lida.
	*/
	int numeroDaLinha() { return linhaAtual; }

	/**
		Use este mtodo para voltar a ler a ltima linha lida. til para
		esquema de lookahead.
	*/
	void devolve(String &) { linhaFoiDevolvida = true; }
};


String StreamNumerado::pegaLinha()
{
	char *ptr;
	if(linhaFoiDevolvida)
	{
		linhaFoiDevolvida = false;
	}
	else
	{
		*buffer = 0;
		s.getline(buffer, TAMBUF);
		++linhaAtual;
	}
	ptr = buffer + strlen(buffer) -1;
	while(ptr > buffer && isspace(*ptr)) // enquanto tem espaos no fim da linha
		*ptr-- = 0;
	ptr = buffer;
	while(isspace(*ptr)) // corta espaos no incio
		++ptr;
	return ptr;
}

enum NivelDeAcesso{PRIVATE, PROTECTED, PUBLIC};


/**
	Classe bsica para os elementos de informao que podem ser retirados
	do fonte C++.
*/
class Elemento
{
public:
	/**
		Nvel de acesso. Um de: PUBLIC, PRIVATE, PROTECTED.
	*/
	NivelDeAcesso nivelDeAcesso;

	/**
		As classes derivadas sobrescrevem este mtodo para extrair
		do fluxo de entrada as informaes que necessitam.
	*/
	virtual void parse(StreamNumerado &s) = 0;

	/**
		As classes derivadas sobrescrevem este mtodo para emitir
		na pgina as suas informaes.
	*/
	virtual void emite(GeradorDeHtml &html) = 0;

	/**
		Contm o texto lido do arquivo.
	*/
	String comentario;

	static String leComentario(StreamNumerado &s);
};

String Elemento::leComentario(StreamNumerado &s)
{
	String result = "";
	for(;;)
	{
		String linha = s.pegaLinha();
		if(linha == "*/") // se for fim de comentrio
			break;
		result += linha + '\n';
	}
	return result;
}

/**
	Representa um membro dentro de uma classe. Contm o comentrio que
	foi feito e a linha que estava logo abaixo do comentrio. Tambm
	armazena o nvel de acesso para aquele membro, se pblico, privado ou
	protegido.
*/
class Membro : public Elemento
{
public:
	/**
		Nvel de acesso. Um de: PUBLIC, PRIVATE, PROTECTED.
	*/
	NivelDeAcesso nivelDeAcesso;

	/**
		Contm a linha de programa que foi comentada.
	*/
	String linha;

	/**
		Extrai as informaes do membro.
	*/
	void parse(StreamNumerado &s);

	/**
		Emite as informaes em formato HTML.
	*/
	void emite(GeradorDeHtml &html);
};

void Membro::emite(GeradorDeHtml &html)
{
	html.paragrafo(linha, true);
	++html;
	html.paragrafo(comentario);
	--html;
}


void Membro::parse(StreamNumerado &s)
{
	linha = s.pegaLinha();

	// Tira '{' da linha, se houver
	String::iterator i = std::find(linha.begin(), linha.end(), '{');
	if(i != linha.end())
		*i = 0;
}

/**
	Contm as informaes referentes a uma classe.
*/
class Classe : public Elemento
{
	/**
		Analisa o cabecalho da declarao da classe. Extrai nome e
		nome das classes bsicas.
	*/
	void parseHeader(StreamNumerado &s);

	/**
		Emite as informaes relativas aos membros da classe que tenham
		este nvel de acesso.
	*/
public:

	/**
		Nome desta classe.
	*/
	String nome;

	/**
		Nome do arquivo onde foi encontrada a declarao desta classe.
	*/
	String nomeDoArquivo;

	/**
		Vetor que contm, temporariamente, os nomes de todas as classes
		das quais esta  derivada.
	*/
	std::vector<String> nomeBasicas;

	/**
		Vetor contendo ponteiros para as classes bsicas desta.
	*/
	std::vector<Classe*> basicas;
	
	/**
		Vetor contendo ponteiros para as classes bsicas desta.
	*/
	std::vector<Classe*> derivadas;

	/**
		Vetor contendo todos os membros comentados desta classe.
	*/
	std::vector<Elemento*> membros;

	/**
		Extrai as informaes relativas  classe.
	*/
	void parse(StreamNumerado &s);

	/**
		Emite apenas o nome desta classe (com link) e opcionalmente
		o das classes derivadas, formando uma rvore de ramificaes.
	*/
	void emiteNome(GeradorDeHtml &html, bool comDerivadas);

	/**
		Emite uma descrio completa da classe,
		com todos os link relevantes.
	*/
	void emite(GeradorDeHtml &html);

	void emiteMembros(GeradorDeHtml &html, NivelDeAcesso nivelDeAcesso);

	/**
		Auxiliar para o processo de ordenao alfabtica.
	*/
	static int compara(const void *lhs, const void *rhs);
};

void Classe::emiteNome(GeradorDeHtml &html, bool comDerivadas)
{
	html.emiteLink(nome, nome);
	if(comDerivadas)
	{
		int i;
		for(i = 0; i < derivadas.size(); ++i)
		{
			++html;
			derivadas[i]->emiteNome(html, comDerivadas);
			--html;
		}
	}
}

void Classe::emiteMembros(GeradorDeHtml &html, NivelDeAcesso nivelDeAcesso)
{
	int quantos = 0, i;
	for(i = 0; i < membros.size(); ++i)
		if(membros[i]->nivelDeAcesso == nivelDeAcesso)
			++quantos;
	if(quantos == 0)
		return;

	const char *nomesDosControlesDeAcesso[] = {"privados", "protegidos", "pblicos"};
	String titulo = String("Membros ") + nomesDosControlesDeAcesso[nivelDeAcesso] + ":";
	html.paragrafo(titulo);
	++html;
	for(i = 0; i < membros.size(); ++i)
	{
		Elemento *ptr = membros[i];
		if(ptr->nivelDeAcesso == nivelDeAcesso)
			ptr->emite(html);
	}
	--html;
}


void Classe::emite(GeradorDeHtml &html)
{
	qsort(derivadas.begin(), derivadas.size(), sizeof(Classe*), Classe::compara); // ordena alfabeticamente
	html.garrafais(nome);
	++html;
	int i;
	for(i = 0; i < basicas.size(); ++i)
	{
		html << " ";
		Classe *ptr = basicas[i];
		while(ptr)
		{
			html.emiteLink(ptr->nome, ptr->nome);
			ptr = (ptr->basicas.size() > 0)? ptr->basicas[0] : 0;
			if(ptr)
				html << ", ";
		}
		html.novaLinha();
	}
	--html;
	html.paragrafo(String("#include \"" + nomeDoArquivo + "\""), true);
	html.paragrafo(comentario);

	emiteMembros(html, PUBLIC);
	emiteMembros(html, PROTECTED);
#ifdef EXIBIR_MEMBROS_PRIVADOS
	emiteMembros(html, PRIVATE);
#endif // EXIBIR_MEMBROS_PRIVADOS

	if(derivadas.size() > 0)
	{
		html.abreParagrafo();
		html << "Veja tambm: ";
		for(i = 0; i < derivadas.size(); ++i)
		{
			Classe *ptr = derivadas[i];
			html.emiteLink(ptr->nome, ptr->nome);
			if(i < derivadas.size() -1)
				html << ", ";
		}
		html.fechaParagrafo();
	}
}

void Classe::parseHeader(StreamNumerado &s)
{
	String linha;
	for(;;)
	{
		linha += s.pegaLinha() + ' ';

		// Tira '{' da linha, se houver
		String::iterator i = std::find(linha.begin(), linha.end(), '{');
		if(i != linha.end())
		{
			*i = 0;
			break;
		}
	}
	const char *ptr = linha.c_str(), *inicio;
	while(*ptr && !isspace(*ptr))  ++ptr;  // pula o que no for espao
	while(*ptr && isspace(*ptr))  ++ptr;  // pula o que for espao
	inicio = ptr;
	while(*ptr && !isspace(*ptr))  ++ptr;  // pula o que no for espao
	nome = String(inicio, ptr - inicio);
	while(ptr && *ptr)
	{
		ptr = strstr(ptr, "public");
		if(ptr == 0)
			break;
		ptr += 6; // pula a palavra
		while(*ptr && isspace(*ptr))  ++ptr;  // pula o que for espao
		inicio = ptr;
		while(*ptr && !isspace(*ptr)) ++ptr;  // pula o que no for espao
		if(ptr[-1] == ',')
			--ptr;
		nomeBasicas.push_back(String(inicio, ptr - inicio));
	}
}


void Classe::parse(StreamNumerado &s)
{
	parseHeader(s);
	NivelDeAcesso nivelDeAcesso = PRIVATE;
	while(s.acabou() == false)
	{
		String linha = s.pegaLinha();
		if(linha == "};")
			break;
		else if(linha == "/**") // se for incio de comentario especial
		{
			String comentario = Elemento::leComentario(s);
			linha = s.pegaLinha(); // lookahead
			s.devolve(linha); // para ficar disponvel para a prxima leitura
			Elemento *elemento = 0;
			if(strncmp(linha.c_str(), "class ", 6) == 0) // se for classe aninhada
			{
				Classe *classe = new Classe;
				classe->nomeDoArquivo = nome;
				elemento = classe;
			}
			else
			{
				elemento = new Membro;
			}
			elemento->nivelDeAcesso = nivelDeAcesso;
			elemento->comentario = comentario;
			elemento->parse(s);
			membros.push_back(elemento);
		}
		else if(linha == "public:")
			nivelDeAcesso = PUBLIC;
		else if(linha == "protected:")
			nivelDeAcesso = PROTECTED;
		else if(linha == "private:")
			nivelDeAcesso = PRIVATE;
	}
}
	
/**
	 o conjunto de todas as classes analisadas.
*/	
class Classes
{
	/**
		L as declaraes de um arquivo.
	*/
	void carregaOArquivo(const char *nome);

	/**
		Resolve os ponteiros internos para as classes.
	*/
	void ligaAsClasses();

	/**
		Emite a pgina de ndice.
	*/
	void geraIndice(String &nomeDoArquivo);

public:
	/**
		Carrega as declaraes de todos os arquivos que batam
		com a mscara.
	*/
	void carregaArquivos(const char *mascara);

	/**
		Gera todas as pginas no diretrio especificado.
	*/
	void geraHtml(const char *diretorio);

protected:
	/**
		Todas as declaraes de classe lidas dos arquivos.
	*/
	std::vector<Classe*> classes;

private:
};


void Classes::ligaAsClasses()
{
	int x, y, z;
	for(x = 0; x < classes.size(); ++x)
	{
		Classe *externa = classes[x];
		std::sort(externa->nomeBasicas.begin(), externa->nomeBasicas.end()); // para ficar em ordem alfabtica
		for(y = 0; y < externa->nomeBasicas.size(); ++y)
		{
			String nomeBasica = externa->nomeBasicas[y];
			for(z = 0; z < classes.size(); ++z)
			{
				Classe *interna = classes[z];
				if(interna->nome == nomeBasica) // se bateu o nome
				{
					externa->basicas.push_back(interna);
					interna->derivadas.push_back(externa);
					break;
				}
			}
			externa->nomeBasicas[y] = ""; // libera memria
		}
	}
}

int Classe::compara(const void *lhs, const void *rhs)
{
	return strcmp((*(Classe**)lhs)->nome.c_str(), (*(Classe**)rhs)->nome.c_str());
}


void Classes::geraIndice(String &nomeDoArquivo)
{
	qsort(classes.begin(), classes.size(), sizeof(Classe*), Classe::compara); // ordena alfabeticamente
	GeradorDeHtml html(String("ndice"), nomeDoArquivo);
	html.garrafais(String("ndice"));
	html.paragrafo(String("Hierarquia:"));
	++html;
	int i;
	for(i = 0; i < classes.size(); ++i)
	{
		if(classes[i]->basicas.size() == 0) // se  mais bsica
		{
			html.novaLinha();
			classes[i]->emiteNome(html, true);
		}
	}
	--html;
	html.paragrafo(String("Lista em ordem alfabtica:"));
	++html;
	for(i = 0; i < classes.size(); ++i)
	{
		classes[i]->emiteNome(html, false);
		html.novaLinha();
	}
	--html;
}


void Classes::geraHtml(const char *diretorio)
{
	ligaAsClasses();

	cout << "Gerando...";
	cout.flush();
	geraIndice(String(diretorio) + "\\index");
	int i;
	for(i = 0; i < classes.size(); ++i)
	{
		GeradorDeHtml html(classes[i]->nome, String(diretorio) + "\\" + classes[i]->nome);
		classes[i]->emite(html);
		html.emiteLink(String("ndice"), String("index"));
	}
	cout << classes.size() +1 << " paginas.\nPara ver o resultado abra index.html";
}

void Classes::carregaOArquivo(const char *nome)
{
	cout << "Carregando " << nome << ' ';
	cout.flush();
	ifstream arq(nome);
	if(!arq.good())
	{
		cerr << "Falha na abertura do arquivo" << endl;
		return;
	}

	StreamNumerado s(arq);
	Classe *classe = 0;
	while(!s.acabou())
	{
		String linha = s.pegaLinha();
		if(linha == "/**") // se for incio de comentrio especial
		{
			String comentario = Elemento::leComentario(s);
			Classe *classe = new Classe;
			classe->comentario = comentario;
			classe->nomeDoArquivo = nome;
			classe->parse(s);
			classes.push_back(classe);
		}
	}
	cout << endl;
}

void Classes::carregaArquivos(const char *nome)
{
	WIN32_FIND_DATA s;
	HANDLE handle = FindFirstFile(nome, &s);
	if(handle == INVALID_HANDLE_VALUE)
	{
		return;
	}
	do
	{
		carregaOArquivo(s.cFileName);
	}while(FindNextFile(handle, &s) != 0);
	FindClose(handle);
}


int main(int argc, char **argv)
{
	if(argc < 2)
	{
		cerr <<
			"GERADOC Copyriht (C) 1998 Pao com Software (by Jucie)\n\n"
			"Gera paginas de documentacao a partir de dados extraidos dos fontes C++.\n"
			"Inspirado no javadoc, da Sun.\n\n"
			"Uso:\n"
			"\t" << *argv << " arquivo.ext [arquivo.ext ...]\n"
			"\nObs.: gera as paginas no diretorio local.\n";
		return 1;
	}

	Classes classes;
	int x;
	for(x = 1; x < argc; ++x)
		classes.carregaArquivos(argv[x]);
	classes.geraHtml(".");
	return 0;
}
